package com.yuandian.dataflow.controller;

import java.time.Instant;
import java.util.List;

import com.alipay.sofa.jraft.Status;
import com.alipay.sofa.jraft.error.RemotingException;
import com.alipay.sofa.jraft.rpc.InvokeCallback;
import com.google.protobuf.Any;
import com.yuandian.dataflow.controller.PacketsProcessor.PacketsRequest;
import com.yuandian.dataflow.proto.msgtype.BacktrackingFlowOuterClass;
import com.yuandian.dataflow.statemachine.StateFactory;
import com.yuandian.dataflow.statemachine.annotations.MasterRegister;
import com.yuandian.dataflow.statemachine.closure.GenericClosure;
import com.yuandian.dataflow.statemachine.master.MasterContext;
import com.yuandian.dataflow.statemachine.master.MasterExecute;
import com.yuandian.dataflow.statemachine.operate.Operate;
import com.yuandian.dataflow.statemachine.operate.Operate.OperateType;
import com.yuandian.dataflow.statemachine.state.State;
import com.yuandian.dataflow.statemachine.state.WorkerState;

import lombok.extern.slf4j.Slf4j;

@Slf4j
@MasterRegister
public class MasterProcessor implements MasterExecute {

    private final int MAX_TASKS = 100;

    @Override
    public void loop(MasterContext cxt) {
        try {
            Thread.sleep(3000);
        } catch (InterruptedException e1) {
            e1.printStackTrace();
            return;
        }

        var alivePeers = List.copyOf(StateFactory.getRaftNode().listAlivePeers());
        log.debug("master({}) execute {}", StateFactory.getServerId(),
                StateFactory.getRaftNode().listAlivePeers());
        if (alivePeers == null) {
            return;
        }

        // 读一致性
        StateFactory.readIndexState(new GenericClosure<State>() {

            @Override
            public void run(Status status) {
                var state = this.getValue();
                // log.debug("masterExecute start {} {}", status, alivePeers);
                alivePeers.forEach((peer) -> {

                    if (state == null) {
                        log.error("readIndexState获取的状态为 {}", state);
                        return;
                    }

                    WorkerState ws = state.getWorkers().get(peer);
                    if (ws == null) {
                        log.error("WorkerState获取的状态为 {}", ws);
                        return;
                    }

                    var canTasks = MAX_TASKS - ws.getTaskQueueSize();
                    log.info("剩余能处理的任务数量[{}] :{}", peer, canTasks);
                    if (canTasks <= 0) {
                        return;
                    }
                    ws.setUpdateAt(Instant.now());
                    ws.setTaskQueueSize(MAX_TASKS);

                    // 模拟发送包的数据到该节点上
                    var request = new PacketsRequest();
                    for (int i = 0; i < canTasks; i++) {
                        var p = Any.pack(BacktrackingFlowOuterClass.BacktrackingFlow
                                .newBuilder()
                                .setTableId(10086)
                                .build());
                        request.getPackets().add(p);
                    }

                    // 先提交 节点的 剩余能处理的任务数量. 然后再处理
                    Operate.CallOperate(
                            new Operate(OperateType.PUT_WORKERSTATE, ws),
                            new GenericClosure<Operate>() {
                                @Override
                                public void run(Status status) {
                                    log.info("PacketsRequest run {}", status);
                                    try {
                                        StateFactory.rpcClientInvokeAsync(peer.getEndpoint(), request,
                                                new InvokeCallback() {
                                                    @Override
                                                    public void complete(Object result, Throwable err) {
                                                        if (err != null) {
                                                            // TODO: 如果错误, 需要让节点恢复任务处理的状态
                                                            log.debug("{}", err);
                                                        }
                                                        log.debug("PacketsRequest: {}", result);
                                                    }
                                                }, 5000);
                                    } catch (InterruptedException | RemotingException e) {
                                        log.info("error send packets {}", e.toString());
                                    }
                                }
                            });
                });
            }

        });
    }

}
